#include <pspsdk.h>
#include <pspkernel.h>
#include <pspctrl.h>
#include <pspiofilemgr.h>
#include <pspdebug.h>
#include <psppower.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
 
PSP_MODULE_INFO("KeyCleaner", 0, 1, 0);
PSP_MAIN_THREAD_ATTR(PSP_THREAD_ATTR_USER);


#define printf pspDebugScreenPrintf

#define VERS    1
#define REVS    1


int ReadKey(int key, char *buffer);
int WriteKey(int key, char *buffer);


int ks[10];
char *psptype;


void wait_release(unsigned int buttons)
{
    SceCtrlData pad;

    sceCtrlReadBufferPositive(&pad, 1);
    while (pad.Buttons & buttons)
    {
        sceKernelDelayThread(100000);
        sceCtrlReadBufferPositive(&pad, 1);
    }
}


unsigned int wait_press(unsigned int buttons)
{
    SceCtrlData pad;

    sceCtrlReadBufferPositive(&pad, 1);
    while (1)
    {
        if (pad.Buttons & buttons)
            return pad.Buttons & buttons;
        sceKernelDelayThread(100000);
        sceCtrlReadBufferPositive(&pad, 1);
    }
    return 0;   /* never reaches here, again, just to suppress warning */
}


int confirm_cancel(void)
{
    SceCtrlData pad;

    while (1)
    {
        sceKernelDelayThread(10000);
        sceCtrlReadBufferPositive(&pad, 1);
        if(pad.Buttons & PSP_CTRL_CROSS)
        {
            wait_release(PSP_CTRL_CROSS);
            return 1;
        }
        if(pad.Buttons & PSP_CTRL_CIRCLE)
        {
            wait_release(PSP_CTRL_CIRCLE);
            return 0;
        }
    }
    return 0;   /* never reaches here, this suppresses a warning */
}


void new_dir(char *dir_name)
{
    char new_name[512];
    int d, err, i;

    d = sceIoDopen(dir_name);

    if (d >= 0)
    {
        /* directory already exists, try to rename it */
        sceIoDclose(d);

        for (i=0; i<10000; i++)
        {
            sprintf(new_name, "%s_%04d", dir_name, i);
            d = sceIoDopen(new_name);
            if (d < 0)
                break; /* directory with this name doesn't exist */
            sceIoDclose(d);
        }
        if (i == 10000)
            printf("\n\n ERROR: Could not rename directory %s!\n\n", dir_name);
        else
        {
            printf("\n Renaming directory %s to %s...", dir_name, new_name);
            err = sceIoRename(dir_name, new_name);
            if (err < 0)
                printf("FAILED! Error code %X\n\n", err);
            else
                printf("PASSED!\n\n");
        }
    }

    sceIoMkdir(dir_name, 0777);
}


void dump_keys(void)
{
    char buffer[512];
    char filepath[32];
    int f, s, currkey;
    int linecnt = 5;

    pspDebugScreenClear();
    printf("\n\n       If keys directory already exists, it will be renamed.\n");
    printf("                O = back, X = dump keys to memstick\n\n");

    if (confirm_cancel())
    {
        new_dir("keys");
        for (currkey=0; currkey<0xfff0; currkey++)
        {
            s = ReadKey(currkey, buffer);
            if (s != 0) continue;
            sprintf(filepath, "keys/0x%04X.bin", currkey);
            f = sceIoOpen(filepath, PSP_O_WRONLY | PSP_O_CREAT, 0777);
            if (f <= 0) continue;
            printf(" Saving key %04X to file %s...", currkey, filepath);
            sceIoWrite(f, buffer, 512);
            sceIoClose(f);
            printf(" done.\n");

            linecnt++;
            if (linecnt >= 28)
            {
                sceKernelDelayThread(1*1000*1000);
                pspDebugScreenClear();
                linecnt = 0;
            }
        }
    }

    sceKernelDelayThread(5*1000*1000);
}


int check_key(int key, char *buffer1)
{
    char buffer2[512];
    char *buffer;
    int s, i;
    int chk[] = {
        0x00004, 0x4272796e,
        0x00005, 0x436c6b67,
        0x00006, 0x4d446472,
        0x00008, 0x4c434470,
        0x00041, 0x0000054c,
        0x00043, 0x72747355,
        0x00045, 0x00010000,    /* PSP-1001 */
        0x00045, 0x00010001,    /* PSP-100? */
        0x00045, 0x00010002,    /* PSP-1004/6 */
        0x00045, 0x00010003,    /* PSP-1000 */
        0x00047, 0x00000009,
        0x10005, 0x436c6b98,    /* Chilly Willy's patch */
        0x20005, 0x436c6867,    /* codes02's patch */
        0x40000, 0x00000000,    /* cleared */
        -1
    };

    printf(" Checking IdStorage key 0x%04X... ", key);

    if (buffer1 == 0)
    {
        s = ReadKey(key, buffer2);
        if (s < 0)
        {
            printf("no key (code 0x%8X)\n", s);
            return s;
        }
        buffer = buffer2;
    }
    else
        buffer = buffer1;

    i = 0;
    while (chk[i] != -1)
    {
        if (chk[i+1] == *((int *)buffer))
            break;
        i += 2;
    }
    if (chk[i] == -1)
    {
        printf("unknown status\n");
        return key;
    }
    if (chk[i] & 0x40000)
    {
        printf("cleared key  - okay\n");
        return chk[i];
    }
    if (key != (chk[i] & 0xffff))
    {
        printf("copy of original key 0x%04X\n", chk[i] & 0xffff);
        return ((chk[i] & 0x0ffff) | 0x80000);
    }
    if (chk[i] & 0x30000)
        printf("patched key  - okay\n");
    else
        printf("original key - okay\n");
    return chk[i];
}


char *check_psp(void)
{
    static char ta079[] = "TA-079/81";
    static char ta082[] = "TA-082/86";

    char buffer[512];

    ReadKey(0x0100, buffer);
    if (buffer[0x03f] != 2)
        return ta079;

    ReadKey(0x0050, buffer);
    if (buffer[0x021] != 2)
        return ta079;

    return ta082;
}


int is_ta082_86(void)
{
    return (strcmp(psptype,"TA-082/86") == 0);
}


int check_state(int *kstatus)
{
    int chk[] = {
        0,  /* original TA-079 */
        0x00004,0x00005,0x00006,-1,0x00041,0x40000,0x00043,0x00045,0x40000,0x80000025,
        1,  /* unpatched PSP */
        0x00004,0x00005,0x00006,-1,0x00041,0x40000,0x00043,0x00045,0x40000,-1,

        2,  /* original soft-downed TA-082/86 */
        0x40000,0x80004,0x80005,-1,0x80006,0x80041,0x40000,0x80047,0x80045,0x40000,
        3,  /* Noobz soft-downed TA-082/86 */
        0x40000,0x80004,0x80005,-1,0x00041,0x40000,0x00043,0x00045,0x40000,0x00047,
        4,  /* codes02 soft-downed TA-082/86 */
        0x00004,0x20005,0x00006,-1,0x00041,0x40000,0x00043,0x00045,0x40000,0x00047,
        5,  /* Chilly Willy soft-downed TA-082/86 */
        0x00004,0x10005,0x00006,-1,0x00041,0x40000,0x00043,0x00045,0x40000,-1,

        6,  /* Chilly Willy soft-downed TA-082/86 with un-fixed key 0x0043 */
        0x00004,0x10005,0x00006,-1,0x00041,0x40000,0x40000,0x00045,0x40000,0x00047,
        7,  /* Chilly Willy soft-downed TA-082/86 with un-fixed key 0x0006 */
        0x00004,0x10005,0x80005,-1,0x00041,0x40000,0x00043,0x00045,0x40000,0x00047,

        8,	/* original soft-down + usbhostfs fix */
        0x40000,0x80004,0x80005,-1,0x00041,0x80041,0x00043,0x80047,0x80045,0x00047,
        9, /* original soft-down + idreset */
        0x00004,0x80004,0x80005,-1,0x00041,0x40000,0x00043,0x40000,0x40000,0x00047,
        10, /* Noobz soft-downed + idcopy */
        0x00004,0x80004,0x80005,-1,0x00041,0x40000,0x00043,0x00045,0x40000,0x00047,

        -1
    };
    char *states[] = {
        "Original TA-079                                    ",
        "Unpatched PSP                                      ",

        "Original soft-downed TA-082/86                     ",
        "NOOBZ soft-downed TA-082/86                        ",
        "codes02 soft-downed TA-082/86                      ",
        "Chilly Willy soft-downed TA-082/86                 ",

        "Chilly Willy soft-downed TA-082/86 - bad key 0x0043",
        "Chilly Willy soft-downed TA-082/86 - bad key 0x0006",

        "Original soft-downed + USBHOSTFS_FIX               ",
        "Original soft-downed + idreset-v7                  ",
        "NOOBZ soft-downed + idcopy                         ",

        "What the h-ll!!                                    "
    };

    int i, j;

    i = 0;
    while (chk[i] != -1)
    {
        for (j=0; j<10; j++)
        {
            if (chk[i+j+1] == -1)
                continue;   /* skip */
            if (chk[i+j+1] != kstatus[j])
            {
                i += 11;
                break;
            }
        }
        if (j == 10)
        {
            printf(" Key state: %s\n\n", states[chk[i]]);
            return chk[i];
        }
    }

    printf(" Unrecognized key state\n\n");
    return -1;
}


void patch_key_0005(void)
{
    char buffer[512];

    printf("\n Are you really sure you wish to do this?\n");
    printf(" Writing the PSP keys has the potential to brick it.\n");
    printf(" Last chance to back out... O = skip, X = go for it!\n\n");
    if (!confirm_cancel())
        return;

    printf(" Patching key 0x0005...");
    ReadKey(0x0005, buffer);
    buffer[1] = (char)0x6b;
    buffer[0] = (char)0x98;
    WriteKey(0x0005, buffer);
    printf("patched!\n\n");

    sceKernelDelayThread(5*1000*1000);
}


int check_dir(char *dname, int keys)
{
    char temp[1024];
    char buffer[512];
    int f;
    int s = -1;
    int ms[10];

    if (keys & 1)
    {
        strcpy(temp, dname);
        strcat(temp, "/0x0004.bin");
        f = sceIoOpen(temp, PSP_O_RDONLY, 0777);
        if (f <= 0)
            ms[0] = 0x80000025;
        else
        {
            sceIoRead(f, buffer, 512);
            sceIoClose(f);
            ms[0] = check_key(4, buffer);
        }
    }

    if (keys & 2)
    {
        strcpy(temp, dname);
        strcat(temp, "/0x0005.bin");
        f = sceIoOpen(temp, PSP_O_RDONLY, 0777);
        if (f <= 0)
            ms[1] = 0x80000025;
        else
        {
            sceIoRead(f, buffer, 512);
            sceIoClose(f);
            ms[1] = check_key(5, buffer);
        }
    }

    if (keys & 4)
    {
        strcpy(temp, dname);
        strcat(temp, "/0x0006.bin");
        f = sceIoOpen(temp, PSP_O_RDONLY, 0777);
        if (f <= 0)
            ms[2] = 0x80000025;
        else
        {
            sceIoRead(f, buffer, 512);
            sceIoClose(f);
            ms[2] = check_key(6, buffer);
        }
    }

    if (keys & 512)
    {
        strcpy(temp, dname);
        strcat(temp, "/0x0008.bin");
        f = sceIoOpen(temp, PSP_O_RDONLY, 0777);
        if (f <= 0)
            ms[3] = 0x80000025;
        else
        {
            sceIoRead(f, buffer, 512);
            sceIoClose(f);
            ms[3] = check_key(8, buffer);
        }
    }

    if (keys & 8)
    {
        strcpy(temp, dname);
        strcat(temp, "/0x0041.bin");
        f = sceIoOpen(temp, PSP_O_RDONLY, 0777);
        if (f <= 0)
            ms[4] = 0x80000025;
        else
        {
            sceIoRead(f, buffer, 512);
            sceIoClose(f);
            ms[4] = check_key(0x41, buffer);
        }
    }

    if (keys & 16)
    {
        strcpy(temp, dname);
        strcat(temp, "/0x0042.bin");
        f = sceIoOpen(temp, PSP_O_RDONLY, 0777);
        if (f <= 0)
            ms[5] = 0x80000025;
        else
        {
            sceIoRead(f, buffer, 512);
            sceIoClose(f);
            ms[5] = check_key(0x42, buffer);
        }
    }

    if (keys & 32)
    {
        strcpy(temp, dname);
        strcat(temp, "/0x0043.bin");
        f = sceIoOpen(temp, PSP_O_RDONLY, 0777);
        if (f <= 0)
            ms[6] = 0x80000025;
        else
        {
            sceIoRead(f, buffer, 512);
            sceIoClose(f);
            ms[6] = check_key(0x43, buffer);
        }
    }

    if (keys & 64)
    {
        strcpy(temp, dname);
        strcat(temp, "/0x0045.bin");
        f = sceIoOpen(temp, PSP_O_RDONLY, 0777);
        if (f <= 0)
            ms[7] = 0x80000025;
        else
        {
            sceIoRead(f, buffer, 512);
            sceIoClose(f);
            ms[7] = check_key(0x45, buffer);
        }
    }

    if (keys & 128)
    {
        strcpy(temp, dname);
        strcat(temp, "/0x0046.bin");
        f = sceIoOpen(temp, PSP_O_RDONLY, 0777);
        if (f <= 0)
            ms[8] = 0x80000025;
        else
        {
            sceIoRead(f, buffer, 512);
            sceIoClose(f);
            ms[8] = check_key(0x46, buffer);
        }
    }

    if (keys & 256)
    {
        strcpy(temp, dname);
        strcat(temp, "/0x0047.bin");
        f = sceIoOpen(temp, PSP_O_RDONLY, 0777);
        if (f <= 0)
            ms[9] = 0x80000025;
        else
        {
            sceIoRead(f, buffer, 512);
            sceIoClose(f);
            ms[9] = check_key(0x47, buffer);
        }
    }

    if (keys == 0x3ff)
        s = check_state(ms);

    return s;
}


void fix_keys_ms(char *dname, int keys)
{
    char temp[1024];
    char buffer[512];
    int f;

    printf("\n Are you really sure you wish to do this?\n");
    printf(" Writing the PSP keys has the potential to brick it.\n");
    printf(" Last chance to back out... O = skip, X = go for it!\n\n");
    if (!confirm_cancel())
        return;

    if (keys & 1)
    {
        strcpy(temp, dname);
        strcat(temp, "/0x0004.bin");
        f = sceIoOpen(temp, PSP_O_RDONLY, 0777);
        if (f > 0)
        {
            printf(" Fixing key 0x0004 from file %s...", temp);
            sceIoRead(f, buffer, 512);
            sceIoClose(f);
            WriteKey(4, buffer);
            printf("fixed!\n");
        }
    }

    if (keys & 2)
    {
        strcpy(temp, dname);
        strcat(temp, "/0x0005.bin");
        f = sceIoOpen(temp, PSP_O_RDONLY, 0777);
        if (f > 0)
        {
            printf(" Fixing key 0x0005 from file %s...", temp);
            sceIoRead(f, buffer, 512);
            sceIoClose(f);
            buffer[1] = (char)0x6b;
            buffer[0] = (char)0x98;     /* make sure key 0x0005 is Chilly Willy'd */
            WriteKey(5, buffer);
            printf("fixed!\n");
        }
    }

    if (keys & 4)
    {
        strcpy(temp, dname);
        strcat(temp, "/0x0006.bin");
        f = sceIoOpen(temp, PSP_O_RDONLY, 0777);
        if (f > 0)
        {
            printf(" Fixing key 0x0006 from file %s...", temp);
            sceIoRead(f, buffer, 512);
            sceIoClose(f);
            WriteKey(6, buffer);
            printf("fixed!\n");
        }
    }

    if (keys & 8)
    {
        strcpy(temp, dname);
        strcat(temp, "/0x0041.bin");
        f = sceIoOpen(temp, PSP_O_RDONLY, 0777);
        if (f > 0)
        {
            printf(" Fixing key 0x0041 from file %s...", temp);
            sceIoRead(f, buffer, 512);
            sceIoClose(f);
            WriteKey(0x41, buffer);
            printf("fixed!\n");
        }
    }

    if (keys & 16)
    {
        strcpy(temp, dname);
        strcat(temp, "/0x0042.bin");
        f = sceIoOpen(temp, PSP_O_RDONLY, 0777);
        if (f > 0)
        {
            printf(" Fixing key 0x0042 from file %s...", temp);
            sceIoRead(f, buffer, 512);
            sceIoClose(f);
            WriteKey(0x42, buffer);
            printf("fixed!\n");
        }
    }

    if (keys & 32)
    {
        strcpy(temp, dname);
        strcat(temp, "/0x0043.bin");
        f = sceIoOpen(temp, PSP_O_RDONLY, 0777);
        if (f > 0)
        {
            printf(" Fixing key 0x0043 from file %s...", temp);
            sceIoRead(f, buffer, 512);
            sceIoClose(f);
            WriteKey(0x43, buffer);
            printf("fixed!\n");
        }
    }

    if (keys & 64)
    {
        strcpy(temp, dname);
        strcat(temp, "/0x0045.bin");
        f = sceIoOpen(temp, PSP_O_RDONLY, 0777);
        if (f > 0)
        {
            printf(" Fixing key 0x0045 from file %s...", temp);
            sceIoRead(f, buffer, 512);
            sceIoClose(f);
            WriteKey(0x45, buffer);
            printf("fixed!\n");
        }
    }

    if (keys & 128)
    {
        strcpy(temp, dname);
        strcat(temp, "/0x0046.bin");
        f = sceIoOpen(temp, PSP_O_RDONLY, 0777);
        if (f > 0)
        {
            printf(" Fixing key 0x0046 from file %s...", temp);
            sceIoRead(f, buffer, 512);
            sceIoClose(f);
            WriteKey(0x46, buffer);
            printf("fixed!\n");
        }
    }

    if (keys & 256)
    {
        strcpy(temp, dname);
        strcat(temp, "/0x0047.bin");
        f = sceIoOpen(temp, PSP_O_RDONLY, 0777);
        if (f > 0)
        {
            printf(" Fixing key 0x0047 from file %s...", temp);
            sceIoRead(f, buffer, 512);
            sceIoClose(f);
            WriteKey(0x47, buffer);
            printf("fixed!\n");
        }
    }

    sceKernelDelayThread(5*1000*1000);
}

void fix_from_ms(int keys)
{
    int dfd, i;
    static SceIoDirent dir;
    char cwd[1024];

    pspDebugScreenInit();

    i = 0;
    getcwd(cwd, 1024);
    dfd = sceIoDopen(cwd);
    if(dfd >= 0)
    {
        memset(&dir, 0, sizeof(dir));
        while(sceIoDread(dfd, &dir) > 0)
        {
            if (strcmp(dir.d_name, ".") == 0 || strcmp(dir.d_name, "..") == 0 || !(dir.d_stat.st_attr & FIO_SO_IFDIR))
            {
                memset(&dir, 0, sizeof(dir));
                continue;
            }
            printf("\n\n Checking %s directory...\n", dir.d_name);
            check_dir(dir.d_name, keys);
            printf("                O = skip this dir, X = use this dir\n");
            if (confirm_cancel())
            {
                fix_keys_ms(dir.d_name, keys);
                return;
            }
            pspDebugScreenInit();
            memset(&dir, 0, sizeof(dir));
        }
    }
}


void fix_orig_sd(void)
{
    char buffer[512];

    printf("\n Are you really sure you wish to do this?\n");
    printf(" Writing the PSP keys has the potential to brick it.\n");
    printf(" Last chance to back out... O = skip, X = go for it!\n\n");
    if (!confirm_cancel())
        return;

    printf(" Fixing key 0x0047...");
    ReadKey(0x0045, buffer);
    WriteKey(0x0047, buffer);
    printf("fixed!\n");

    printf(" Fixing key 0x0045...");
    ReadKey(0x0046, buffer);
    WriteKey(0x0045, buffer);
    printf("fixed!\n");

    printf(" Fixing key 0x0046...");
    ReadKey(0x0004, buffer);
    WriteKey(0x0046, buffer);
    printf("fixed!\n");

    printf(" Fixing key 0x0004...");
    ReadKey(0x0005, buffer);
    WriteKey(0x0004, buffer);
    printf("fixed!\n");

    printf(" Fixing key 0x0005...");
    ReadKey(0x0006, buffer);
    buffer[1] = (char)0x6b;
    buffer[0] = (char)0x98;     /* make sure key 0x0005 is Chilly Willy'd */
    WriteKey(0x0005, buffer);
    printf("fixed!\n");

    printf(" Fixing key 0x0006...");
    ReadKey(0x0041, buffer);
    WriteKey(0x0006, buffer);
    printf("fixed!\n");

    printf(" Fixing key 0x0041...");
    ReadKey(0x0042, buffer);
    WriteKey(0x0041, buffer);
    printf("fixed!\n");

    printf(" Fixing key 0x0042...");
    ReadKey(0x0043, buffer);
    WriteKey(0x0042, buffer);
    printf("fixed!\n");

    printf(" All keys but 0x0043 fixed\n");
    sceKernelDelayThread(5*1000*1000);
}


void fix_noobz_sd(void)
{
    char buffer[512];

    printf("\n Are you really sure you wish to do this?\n");
    printf(" Writing the PSP keys has the potential to brick it.\n");
    printf(" Last chance to back out... O = skip, X = go for it!\n\n");
    if (!confirm_cancel())
        return;

    printf(" Fixing key 0x0004...");
    ReadKey(0x0005, buffer);
    WriteKey(0x0004, buffer);
    printf("fixed!\n");

    printf(" Fixing key 0x0005...");
    ReadKey(0x0006, buffer);
    buffer[1] = (char)0x6b;
    buffer[0] = (char)0x98;     /* make sure key 0x0005 is Chilly Willy'd */
    WriteKey(0x0005, buffer);
    printf("fixed!\n");

    printf(" All keys but 0x0006 fixed\n");
    sceKernelDelayThread(5*1000*1000);
}


void fix_idreset_sd(void)
{
    char buffer[512];
    unsigned int b, r;

    printf("\n Are you really sure you wish to do this?\n");
    printf(" Writing the PSP keys has the potential to brick it.\n");
    printf(" Last chance to back out... O = skip, X = go for it!\n\n");
    if (!confirm_cancel())
        return;

    printf(" Fixing key 0x0005...");
    ReadKey(0x0006, buffer);
    buffer[1] = (char)0x6b;
    buffer[0] = (char)0x98;     /* make sure key 0x0005 is Chilly Willy'd */
    WriteKey(0x0005, buffer);
    printf("fixed!\n");

    printf(" What model is your PSP? O = PSP-1000, X = PSP-1001, %c = PSP-1004/6\n", 0xc8);
    b = wait_press(PSP_CTRL_SQUARE|PSP_CTRL_CROSS|PSP_CTRL_CIRCLE);
    wait_release(PSP_CTRL_SQUARE|PSP_CTRL_CROSS|PSP_CTRL_CIRCLE);
    switch (b)
    {
        case PSP_CTRL_CROSS:
            r = 0;
            break;
        case PSP_CTRL_SQUARE:
            r = 2;
            break;
        case PSP_CTRL_CIRCLE:
            r = 3;
            break;
        default:
            r = 0;
            break;
    }
    printf(" Fixing key 0x0045...");
    memset(buffer, 0, 512);
    buffer[0] = (char)r;
    buffer[2] = (char)0x01;
    WriteKey(0x0045, buffer);
    printf("fixed!\n");

    printf(" All keys but 0x0006 fixed\n");
    sceKernelDelayThread(5*1000*1000);
}


void fix_keys(void)
{
    int s;
    unsigned int b;

    pspDebugScreenInit();
    printf("\n\n");
    s = check_state(ks);

    if (is_ta082_86())
        switch (s)
        {
            case 1:
                printf(" PSP is an unpatched TA-082/86\n");
                printf(" Recommended action: patch key 0x0005\n\n");
                printf("        O = return to previous screen, X = patch key 0x0005\n");
                if (confirm_cancel())
                    patch_key_0005();
                break;
            case 2:
                printf(" PSP is an original soft-downed TA-082/86... keys need fixing\n");
                printf(" Recommended action: restore keys from memstick\n\n");
                printf(" O = previous screen, X = fix keys using memstick, %c = fix w/o files\n", 0xc8);
                b = wait_press(PSP_CTRL_SQUARE|PSP_CTRL_CROSS|PSP_CTRL_CIRCLE);
                wait_release(PSP_CTRL_SQUARE|PSP_CTRL_CROSS|PSP_CTRL_CIRCLE);
                switch (b)
                {
                    case PSP_CTRL_CROSS:
                        fix_from_ms(0x3ff);
                        break;
                    case PSP_CTRL_SQUARE:
                        fix_orig_sd();
                        break;
                    case PSP_CTRL_CIRCLE:
                        break;
                }
                break;
            case 3:
                printf(" PSP is a NOOBZ soft-downed TA-082/86... keys need fixing\n");
                printf(" Recommended action: restore keys from memstick\n\n");
                printf(" O = previous screen, X = fix keys using memstick, %c = fix w/o files\n", 0xc8);
                b = wait_press(PSP_CTRL_SQUARE|PSP_CTRL_CROSS|PSP_CTRL_CIRCLE);
                wait_release(PSP_CTRL_SQUARE|PSP_CTRL_CROSS|PSP_CTRL_CIRCLE);
                switch (b)
                {
                    case PSP_CTRL_CROSS:
                        fix_from_ms(7);
                        break;
                    case PSP_CTRL_SQUARE:
                        fix_noobz_sd();
                        break;
                    case PSP_CTRL_CIRCLE:
                        break;
                }
                break;
            case 4:
                printf(" PSP is a codes02 patched TA-082/86\n");
                printf(" Recommended action: re-patch key 0x0005\n\n");
                printf("       O = return to previous screen, X = re-patch key 0x0005\n");
                if (confirm_cancel())
                    patch_key_0005();
                break;
            case 5:
                printf(" PSP is a Chilly Willy patched TA-082/86\n");
                printf(" Recommended action: none\n\n");
                printf("               O = return to previous screen\n");
                confirm_cancel();
                break;
            case 6:
                printf(" PSP is a Chilly Willy patched TA-082/86 with bad key 0x0043\n");
                printf(" Recommended action: fix key 0x0043 using memstick file\n\n");
                printf("      O = return to previous screen, X = fix key using memstick\n");
                if (confirm_cancel())
                    fix_from_ms(32);
                break;
            case 7:
                printf(" PSP is a Chilly Willy patched TA-082/86 with bad key 0x0006\n");
                printf(" Recommended action: fix key 0x0006 using memstick file\n\n");
                printf("      O = return to previous screen, X = fix key using memstick\n");
                if (confirm_cancel())
                    fix_from_ms(4);
                break;
            case 8:
                printf(" PSP is an original soft-downed TA-082/86 + USBHOSTFS_FIX\n");
                printf(" Recommended action: restore keys from memstick\n\n");
                printf("     O = return to previous screen, X = fix keys using memstick\n");
                if (confirm_cancel())
                    fix_from_ms(0x3ff);
                break;
            case 9:
                printf(" PSP is an original soft-downed TA-082/86 + idreset-v7\n");
                printf(" Recommended action: restore keys from memstick\n\n");
                printf(" O = previous screen, X = fix keys using memstick, %c = fix w/o files\n", 0xc8);
                b = wait_press(PSP_CTRL_SQUARE|PSP_CTRL_CROSS|PSP_CTRL_CIRCLE);
                wait_release(PSP_CTRL_SQUARE|PSP_CTRL_CROSS|PSP_CTRL_CIRCLE);
                switch (b)
                {
                    case PSP_CTRL_CROSS:
                        fix_from_ms(0x3ff);
                        break;
                    case PSP_CTRL_SQUARE:
                        fix_idreset_sd();
                        break;
                    case PSP_CTRL_CIRCLE:
                        break;
                }
                break;
            case 10:
                printf(" PSP is a NOOBZ soft-downed TA-082/86 + idcopy\n");
                printf(" Recommended action: restore keys from memstick\n\n");
                printf(" O = previous screen, X = fix keys using memstick, %c = fix w/o files\n", 0xc8);
                b = wait_press(PSP_CTRL_SQUARE|PSP_CTRL_CROSS|PSP_CTRL_CIRCLE);
                wait_release(PSP_CTRL_SQUARE|PSP_CTRL_CROSS|PSP_CTRL_CIRCLE);
                switch (b)
                {
                    case PSP_CTRL_CROSS:
                        fix_from_ms(7);
                        break;
                    case PSP_CTRL_SQUARE:
                        fix_noobz_sd();
                        break;
                    case PSP_CTRL_CIRCLE:
                        break;
                }
                break;
            default:
                printf(" PSP is a TA-082/86... keys may need fixing\n");
                printf(" Recommended action: restore keys from memstick\n\n");
                printf("     O = return to previous screen, X = fix keys using memstick\n");
                if (confirm_cancel())
                    fix_from_ms(0x3ff);
                break;
        }
    else
        switch (s)
        {
            case 0:
            case 1:
                printf(" PSP is a TA-079/81... keys check as fine\n\n");
                printf(" Recommended action: none\n\n");
                printf("               O = return to previous screen\n");
                confirm_cancel();
                break;
            default:
                printf(" PSP is a TA-079/81... keys may need fixing\n");
                printf(" Recommended action: restore keys from memstick\n\n");
                printf("     O = return to previous screen, X = fix keys using memstick\n");
                if (confirm_cancel())
                    fix_from_ms(0x3ff);
                break;
        }

    return;
}


int main(void)
{
    unsigned int b;

    pspDebugScreenInit();
    sceCtrlSetSamplingCycle(0);
    sceCtrlSetSamplingMode(PSP_CTRL_MODE_DIGITAL);

    if (scePowerGetBatteryLifePercent() < 75)
    {
        printf("\n\n Battery has to be at least at 75%%.\n");
        sceKernelDelayThread(5*1000*1000);
        sceKernelExitGame();
    }

    SceUID mod = pspSdkLoadStartModule("idstorage.prx", PSP_MEMORY_PARTITION_KERNEL);
    if (mod < 0)
    {
        printf(" Error 0x%08X loading/starting idstorage.prx.\n", mod);
        sceKernelDelayThread(5*1000*1000);
        sceKernelExitGame();
    }

mloop:
    pspDebugScreenClear();
    printf("\n\n                         Key Cleaner v%d.%d\n", VERS, REVS);
    printf("                         by  Chilly Willy\n\n");
    sceKernelDelayThread(2*1000*1000);

    ks[0] = check_key(0x0004, 0);
    ks[1] = check_key(0x0005, 0);
    ks[2] = check_key(0x0006, 0);
    ks[3] = check_key(0x0008, 0);
    ks[4] = check_key(0x0041, 0);
    ks[5] = check_key(0x0042, 0);
    ks[6] = check_key(0x0043, 0);
    ks[7] = check_key(0x0045, 0);
    ks[8] = check_key(0x0046, 0);
    ks[9] = check_key(0x0047, 0);

    psptype = check_psp();
    printf("\n PSP is a %s\n\n", psptype);

    check_state(ks);

    printf("               O = exit, X = fix keys, %c = dump keys\n\n", 0xc8);

    /* check buttons */
    while (1)
    {
        b = wait_press(PSP_CTRL_SQUARE|PSP_CTRL_CROSS|PSP_CTRL_CIRCLE|PSP_CTRL_TRIANGLE);
        wait_release(PSP_CTRL_SQUARE|PSP_CTRL_CROSS|PSP_CTRL_CIRCLE|PSP_CTRL_TRIANGLE);
        switch (b)
        {
            case PSP_CTRL_CROSS:
                fix_keys();
                goto mloop;
                break;
            case PSP_CTRL_SQUARE:
                dump_keys();
                goto mloop;
                break;
            case PSP_CTRL_CIRCLE:
                printf("\n\n Exiting application. Please wait for XMB to reload.\n");
                sceKernelDelayThread(2*1000*1000);
                sceKernelExitGame();
                break;
        }
    }

    return 0;   /* never reaches here, again, just to suppress warning */
}
